from pytest_cases import fixture_ref, parametrize_plus

from carnival import cmd, global_context


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_daemon_reload(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            cmd.systemd.daemon_reload()


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_start(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            cmd.systemd.start("nginx", reload_daemon=True)


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_stop(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            cmd.systemd.stop("nginx", reload_daemon=True)


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_restart(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            cmd.systemd.restart("nginx")


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_enable(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            cmd.systemd.enable("nginx", reload_daemon=True, start_now=True)


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_disable(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            cmd.systemd.disable("nginx", reload_daemon=True, stop_now=True)



from carnival.step import Step
from carnival.host import Host
from carnival.task import Task, SimpleTask
from carnival import secrets_manager
from carnival import cmd
from carnival.utils import log


__all__ = [
    'Step',
    'Host',
    'Task', 'SimpleTask',
    'secrets_manager',
    'cmd',
    'log',
]



from pytest_cases import fixture_ref, parametrize_plus

from carnival import cmd, global_context


@parametrize_plus('host', [
    fixture_ref('local_host'),
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_is_dir_exists(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            assert cmd.fs.is_dir_exists("/etc")
            assert cmd.fs.is_dir_exists("/bin")


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_mkdirs(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            assert cmd.fs.is_dir_exists("/tmp/.carnivaltestdir1") is False
            assert cmd.fs.is_dir_exists("/tmp/.carnivaltestdir2") is False

            cmd.fs.mkdirs("/tmp/.carnivaltestdir1", "/tmp/.carnivaltestdir2")

            assert cmd.fs.is_dir_exists("/tmp/.carnivaltestdir1") is True
            assert cmd.fs.is_dir_exists("/tmp/.carnivaltestdir2") is True

            cmd.cli.run(f"rm -rf /tmp/.carnivaltestdir1 /tmp/.carnivaltestdir2")



from invoke import Context

from carnival import global_context, Host


def test_global_context():
    h = Host("local")
    assert global_context.host is None
    assert global_context.conn is None

    with global_context.SetContext(h):
        assert global_context.host == h
        assert global_context.conn is not None
        assert isinstance(global_context.conn, Context)

    assert global_context.host is None
    assert global_context.conn is None



from typing import Type

import pytest

from carnival import Step, Host


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "slow: slow tests"
    )


@pytest.fixture(scope="function")
def noop_step_class() -> Type[Step]:
    class NoopStep(Step):
        def run(self):
            pass
    return NoopStep


@pytest.fixture(scope="function")
def noop_step(noop_step_class) -> Step:
    return noop_step_class()


@pytest.fixture(scope="function")
def noop_step_context(noop_step_class) -> Step:
    return noop_step_class(additional="context")


@pytest.fixture(scope='function')
def suspend_capture(pytestconfig):
    # https://github.com/pytest-dev/pytest/issues/1599
    # Connection local need turn off capturing
    class suspend_guard:
        def __init__(self):
            self.capmanager = pytestconfig.pluginmanager.getplugin('capturemanager')

        def __enter__(self):
            self.capmanager.suspend_global_capture(in_=True)
            pass

        def __exit__(self, _1, _2, _3):
            self.capmanager.resume_global_capture()

    yield suspend_guard()


@pytest.fixture(scope="function")
def local_host():
    return Host("local")


@pytest.fixture(scope="function")
def ubuntu_ssh_host():
    return Host("127.0.0.1", ssh_user="root", ssh_password="secret", ssh_port=22222)


@pytest.fixture(scope="function")
def centos_ssh_host():
    return Host("127.0.0.1", ssh_user="root", ssh_password="secret", ssh_port=22223)



from patchwork import files  # type:ignore

from carnival import cmd

from carnival import global_context


def mkdirs(*dirs: str):
    return [cmd.cli.run(f"mkdir -p {x}", hide=True) for x in dirs]


def is_dir_exists(dir_path: str) -> bool:
    return cmd.cli.run(f"test -d {dir_path}", warn=True, hide=True).ok


def is_file_contains(filename, text, exact=False, escape=True) -> bool:
    # https://fabric-patchwork.readthedocs.io/en/latest/api/files.html#patchwork.files.contains
    return files.contains(global_context.conn, runner=global_context.conn.run, filename=filename, text=text, exact=exact, escape=escape)


def is_file_exists(path) -> bool:
    # https://fabric-patchwork.readthedocs.io/en/latest/api/files.html#patchwork.files.exists
    return files.exists(global_context.conn, runner=global_context.conn.run, path=path)


def ensure_dir_exists(path, user=None, group=None, mode=None) -> None:
    # https://fabric-patchwork.readthedocs.io/en/latest/api/files.html#patchwork.files.directory
    return files.directory(global_context.conn, runner=global_context.conn.run, path=path, user=user, group=group, mode=mode)



from carnival import utils, Host, global_context


def test_log(capsys):
    with global_context.SetContext(Host("local")):
        utils.log("Hellotest")

        captured = capsys.readouterr()
        assert captured.out == "💃💃💃 local> Hellotest\n"



from typing import Union

from fabric import Connection  # type: ignore
from invoke import Context  # type: ignore

LOCAL_ADDRS = [
    'local',
    'localhost',
]


class Host:
    """
    Объект, представляющий единицу оборудования.

    Carnival не предоставляет никаких сложных абстракций для работы с группами хостов,
    подразумевая что вы будете использовать встроенные коллекции python и организуете
    работу так, как будет удобно для вашей задачи.

    >>> class SetupFrontend(Task):
    >>>    def run(self, **kwargs):
    >>>        self.step(Frontend(), Host("1.2.3.4", packages=["htop", ]))

    В более сложных, создать списки в файле `inventory.py`

    >>> # inventory.py
    >>> frontends = [
    >>>     Host("1.2.3.4"),
    >>>     Host("1.2.3.5"),
    >>> ]

    >>> # carnival_tasks.py
    >>> import inventory as i
    >>> class SetupFrontend(Task):
    >>>    def run(self, **kwargs):
    >>>        self.step(Frontend(), i.frontends)


    """
    def __init__(
        self,
        addr: str,
        ssh_user: str = None, ssh_password: str = None, ssh_port=22,
        ssh_connect_timeout: int = 10,
        **context
     ):
        """
        В простом случае, можно передавать хосты прямо в коде файла `carnival_tasks.py`.

        :param addr: Адрес сервера
        :param ssh_user: Пользователь SSH
        :param ssh_password: Пароль SSH
        :param ssh_port: SSH порт
        :param ssh_connect_timeout: SSH таймаут соединения
        :param context: Контекст хоста
        """
        self.addr = addr
        self.ssh_port = ssh_port
        self.context = context
        self.ssh_user = ssh_user
        self.ssh_password = ssh_password
        self.ssh_connect_timeout = ssh_connect_timeout

    def is_connection_local(self) -> bool:
        """
        Check if host's connection is local
        """
        return self.host.lower() in LOCAL_ADDRS

    def connect(self) -> Union[Connection, Context]:
        if self.is_connection_local():
            # Host is local machine
            return Context()
        else:
            # Host is remote ssh machine
            return Connection(
                host=self.addr,
                port=self.ssh_port,
                user=self.ssh_user,
                connect_timeout=self.ssh_connect_timeout,
                connect_kwargs={
                    'password': self.ssh_password,
                }
            )

    @property
    def host(self) -> str:
        """
        Remove user and port parts, return just address
        """

        h = self.addr
        if '@' in self.addr:
            h = h.split("@", maxsplit=1)[1]

        if ':' in self.addr:
            h = h.split(":", maxsplit=1)[0]

        return h

    def __str__(self):
        return f"🖥 {self.host}"

    def __hash__(self):
        return hash(self.addr)



from jinja2 import Environment, DictLoader

from carnival import templates


def test_render(mocker):
    mocker.patch(
        'carnival.templates.j2_env',
        new=Environment(loader=DictLoader({"index.html": "Hello: {{ name }}"})),
    )
    rendered = templates.render("index.html", name="world")
    assert rendered == "Hello: world"



from pytest_cases import fixture_ref, parametrize_plus

from carnival import cmd, global_context

KEY = "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABAQCzlL9Wo8ywEFXSvMJ8FYmxP6HHHMDTyYAWwM3AOtsc96DcYVQIJ5VsydZf5/4NWuq55MqnzdnGB2IfjQvOrW4JEn0cI5UFTvAG4PkfYZb00Hbvwho8JsSAwChvWU6IuhgiiUBofKSMMifKg+pEJ0dLjks2GUcfxeBwbNnAgxsBvY6BCXRfezIddPlqyfWfnftqnafIFvuiRFB1DeeBr24kik/550MaieQpJ848+MgIeVCjko4NPPLssJ/1jhGEHOTlGJpWKGDqQK+QBaOQZh7JB7ehTK+pwIFHbUaeAkr66iVYJuC05iA7ot9FZX8XGkxgmhlnaFHNf0l8ynosanqt example@laptop"  # noqa


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_ssh_authorized_keys_list(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            assert cmd.system.ssh_authorized_keys_list() == []
            cmd.system.ssh_authorized_keys_add(KEY)
            keys = cmd.system.ssh_authorized_keys_list()
            assert len(keys) == 1
            assert KEY in keys

            # Check not added second time
            cmd.system.ssh_authorized_keys_add(KEY)
            keys = cmd.system.ssh_authorized_keys_list()
            assert len(keys) == 1
            assert KEY in keys

            cmd.cli.run(f"rm ~/.ssh/authorized_keys", hide=True)


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_ssh_authorized_keys_ensure(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            assert cmd.system.ssh_authorized_keys_list() == []
            cmd.system.ssh_authorized_keys_ensure(KEY)
            keys = cmd.system.ssh_authorized_keys_list()
            assert len(keys) == 1
            assert KEY in keys

            # Check not added second time
            cmd.system.ssh_authorized_keys_ensure(KEY)
            keys = cmd.system.ssh_authorized_keys_list()
            assert len(keys) == 1
            assert KEY in keys

            cmd.cli.run(f"rm ~/.ssh/authorized_keys", hide=True)


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_get_current_user_name(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            assert cmd.system.get_current_user_name() == 'root'


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_get_current_user_id(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            assert cmd.system.get_current_user_id() == 0


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_is_current_user_root(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            assert cmd.system.is_current_user_root() is True



import warnings
import getpass
import os
from typing import Dict, Any

from carnival.secrets_manager.base import SecretGetter


secrets_storage: Dict[str, Any] = {}


def secret(var_name: str, secret_get_method: SecretGetter):
    if var_name in secrets_storage.keys():
        warnings.warn(f"Secret {var_name} already defined. Skipping.")

    secrets_storage[var_name] = secret_get_method.get_secret(var_name)


class SecretGetError(Exception):
    pass


class Static(SecretGetter):
    def __init__(self, value: str):
        self._value = value

    def get_secret(self, var_name: str):
        return self._value


class FromCli(SecretGetter):
    """
    Ask in cli
    """
    def get_secret(self, var_name: str):
        return getpass.getpass(f"{var_name}> ")


class FromEnv(SecretGetter):
    """
    Get from environment
    """
    def __init__(self, default=None, required=False):
        self.default = default
        self.required = required

    def get_secret(self, var_name: str):
        res = os.getenv(var_name, None)
        if res is None:
            if self.required is True:
                raise SecretGetError(f"{var_name} is not set in environment")
            return self.default
        return res



from carnival import Task


class Help(Task):
    """
    Показать список доступных задач
    """
    module_name = ""

    def run(self):
        from carnival.cli import task_types
        task_list = list(task_types.keys())
        task_list.sort()

        for task in task_list:
            print(f" * {task}")



from jinja2 import Environment, DictLoader

from pytest_cases import fixture_ref, parametrize_plus

from carnival import cmd, global_context


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_put_template(suspend_capture, host, mocker):
    with suspend_capture:
        with global_context.SetContext(host):
            mocker.patch(
                'carnival.templates.j2_env',
                new=Environment(loader=DictLoader({"index.html": "Hello: {{ name }}"})),
            )
            assert cmd.fs.is_file_exists("/index") is False
            cmd.transfer.put_template("index.html", "/index")
            assert cmd.fs.is_file_exists("/index") is True
            cmd.cli.run("rm /index")


@parametrize_plus('host', [
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_rsync(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            cmd.system.ssh_copy_id()
            assert cmd.fs.is_dir_exists("/docs") is False
            cmd.transfer.rsync("./docs", "/")
            assert cmd.fs.is_dir_exists("/docs") is True
            cmd.cli.run("rm -rf /docs")



from typing import Dict, Any
import importlib
import pkgutil


def discover_plugins() -> Dict[str, Any]:
    """
    Get modules, name starting with carnival_
    """

    discovered_plugins = {
        name: importlib.import_module(name)
        for finder, name, ispkg in pkgutil.iter_modules() if name.startswith('carnival_') and name != 'carnival_tasks'
    }
    return discovered_plugins



from pytest_cases import fixture_ref, parametrize_plus

from carnival import cmd, global_context


@parametrize_plus('host', [
    fixture_ref('local_host'),
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_run(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            result = cmd.cli.run("ls -1 /", hide=True)
            assert result.ok is True

            # Check response looks like root fs
            root_files = result.stdout.split("\n")
            assert 'bin' in root_files
            assert 'etc' in root_files
            assert 'usr' in root_files
            assert 'tmp' in root_files
            assert 'sbin' in root_files


@parametrize_plus('host', [
    fixture_ref('local_host'),
    fixture_ref('ubuntu_ssh_host'),
    fixture_ref('centos_ssh_host'),
])
def test_pty(suspend_capture, host):
    with suspend_capture:
        with global_context.SetContext(host):
            result = cmd.cli.pty("ls -1 / | grep bin", hide=True)
            assert result.ok is True

            # Check response looks like root fs, filtered 'bin'
            root_files = result.stdout.split("\n")
            assert 'bin\r' in root_files
            assert 'sbin\r' in root_files



import pytest

from carnival import Host
from carnival.task import _underscore, Task, SimpleTask


def test_underscore():
    assert _underscore("TestKlass") == 'test_klass'


def test_task_name():
    class DryTask(Task):
        pass

    t = DryTask(True)
    assert t.get_name() == "dry_task"

    class DryNameTask(Task):
        name = "nametask"

    t = DryNameTask(True)
    assert t.get_name() == "nametask"


def test_task_dry_run(noop_step, mocker):
    spy = mocker.spy(noop_step, 'run')

    class DryTask(Task):
        def run(self):
            self.step(noop_step, Host("local"))
    t = DryTask(True)

    t.run()
    spy.assert_not_called()


def test_task(noop_step, mocker):
    spy = mocker.spy(noop_step, 'run')

    with pytest.raises(NotImplementedError):
        Task(False).run()

    class DryTask(Task):
        def run(self):
            self.step(noop_step, Host("local"))
            self.step([noop_step, ], [Host("local"), ])
    t = DryTask(False)

    t.run()
    spy.assert_called()


def test_simple_task(noop_step, mocker):
    spy = mocker.spy(noop_step, 'run')

    with pytest.raises(NotImplementedError):
        Task(False).run()

    class DryTask(SimpleTask):
        hosts = [Host("local"), ]
        steps = [noop_step, ]
    t = DryTask(False)

    t.run()
    spy.assert_called()



from carnival import Host


def test_host_create():
    assert Host("root@1.2.3.4:22").host == "1.2.3.4"
    assert Host("1.2.3.4:22").host == "1.2.3.4"
    assert Host("1.2.3.4").host == "1.2.3.4"
    assert Host("abc.example:22").host == "abc.example"
    assert Host("user@abc.example:22").host == "abc.example"

    assert Host("local").is_connection_local() is True
    assert Host("localhost").is_connection_local() is True
    assert Host("user@localhost").is_connection_local() is True

    assert Host("1.2.3.4").is_connection_local() is False
    assert Host("root@1.2.3.4").is_connection_local() is False
    assert Host("root@example.com").is_connection_local() is False

    #  TODO test connection


def test_host_hash():
    h1 = Host("1.2.3.4")
    h2 = Host("1.2.3.4")

    assert h1.__hash__() == h2.__hash__()



from typing import Optional, List

from carnival import cmd
from carnival.utils import log


def get_pkg_versions(pkgname: str) -> List[str]:
    versions = []
    result = cmd.cli.run(f"apt-cache madison {pkgname}", hide=True, warn=True)
    if result.ok is False:
        return []

    for line in result.stdout.strip().split("\n"):
        n, ver, r = line.split("|")
        versions.append(ver.strip())
    return versions


def get_installed_version(pkgname: str) -> Optional[str]:
    """
    Get installed package version
    Returns None if package not installed
    """
    result = cmd.cli.run(f"dpkg -l {pkgname}", hide=True, warn=True)
    if result.ok is False:
        return None
    installed, pkgn, ver, arch, *desc = result.stdout.strip().split("\n")[-1].split()
    if installed != 'ii':
        return None
    return ver.strip()


def is_pkg_installed(pkgname: str, version=None) -> bool:
    """
    Check is package installed?
    If version not specified - check any version
    """

    pkgver = get_installed_version(pkgname)
    if version is None and pkgver is not None:
        return True

    if version is not None and pkgver == version:
        return True

    return False


def force_install(pkgname, version=None, update=False, hide=False):
    """
    Install apt package
    """
    if version:
        pkgname = f"{pkgname}={version}"

    if update:
        cmd.cli.run("sudo apt-get update", pty=True, hide=hide)

    cmd.cli.run(f"sudo apt-get install -y {pkgname}", pty=True, hide=hide)


def install(pkgname, version=None, update=True, hide=False) -> bool:
    """
    Install apt package if not installed
    Returns true if installed, false if was already installed
    """
    if is_pkg_installed(pkgname, version):
        if version:
            if not hide:
                log(f"{pkgname}={version} already installed")
        else:
            if not hide:
                log(f"{pkgname} already installed")
        return False
    force_install(pkgname=pkgname, version=version, update=update, hide=hide)
    return True


def install_multiple(*pkg_names: str, update=True, hide=False) -> bool:
    if all([is_pkg_installed(x) for x in pkg_names]):
        if not hide:
            log(f"{','.join(pkg_names)} already installed")
        return False

    if update:
        cmd.cli.run("sudo apt-get update", pty=True, hide=hide)

    for pkg in pkg_names:
        install(pkg, update=False, hide=hide)
    return True


def remove(*pkg_names: str, hide=False):
    assert pkg_names, "pkg_names is empty"
    cmd.cli.run(f"sudo apt-get remove --auto-remove -y {' '.join(pkg_names)}", hide=hide)



from typing import List
import os

from carnival import cmd


def set_password(username: str, password: str):
    cmd.cli.pty(f"echo '{username}:{password}' | chpasswd", hide=True)


def ssh_authorized_keys_add(ssh_key: str, keys_file=".ssh/authorized_keys"):
    ssh_key = ssh_key.strip()

    cmd.cli.run("mkdir -p ~/.ssh")
    cmd.cli.run("chmod 700 ~/.ssh")
    cmd.cli.run(f"touch {keys_file}")

    if not cmd.fs.is_file_contains(keys_file, ssh_key, escape=True):
        cmd.cli.run(f"echo '{ssh_key}' >> {keys_file}")


def ssh_authorized_keys_list() -> List[str]:
    if cmd.fs.is_file_exists("~/.ssh/authorized_keys") is False:
        return []

    return cmd.cli.run("cat ~/.ssh/authorized_keys", hide=True).stdout.strip().split("\n")


def ssh_authorized_keys_ensure(*ssh_keys: str) -> None:
    for ssh_key in ssh_keys:
        ssh_authorized_keys_add(ssh_key)


def ssh_copy_id(pubkey_file="~/.ssh/id_rsa.pub") -> None:
    ssh_authorized_keys_add(open(os.path.expanduser(pubkey_file)).read().strip())


def get_current_user_name() -> str:
    return cmd.cli.run("id -u -n", hide=True).stdout.strip()


def get_current_user_id() -> int:
    return int(cmd.cli.run("id -u", hide=True).stdout.strip())


def is_current_user_root() -> bool:
    return get_current_user_id() == 0



import os
import sys
import abc
from typing import Type, Dict, Set

from carnival.task import Task


def task_subclasses(cls) -> Set[Type]:
    # Get subclasses of task, which not abstract

    subclasses = set()
    for sc in cls.__subclasses__():
        # Skip if last MRO base is ABC
        if abc.ABC != sc.__mro__[1]:
            subclasses.add(sc)

        subclasses.update(task_subclasses(sc))

    return subclasses


def get_task_full_name(carnival_tasks_module: str, task_class: Type[Task]) -> str:
    task_name = task_class.get_name()

    task_mod = task_class.module_name
    if task_mod is None:
        task_mod = task_class.__module__

    if task_mod == "":
        return task_name

    task_full_name = f"{task_mod}.{task_name}"

    if task_full_name.startswith(f"{carnival_tasks_module}"):
        task_full_name = task_full_name[len(carnival_tasks_module) + 1:]

    return task_full_name


def import_tasks_file(carnival_tasks_module: str, silent: bool) -> None:
    try:
        __import__(carnival_tasks_module)
    except (ModuleNotFoundError, FileNotFoundError) as ex:
        if not silent:
            print(f"Cannot import {carnival_tasks_module}: {ex}", file=sys.stderr)


def get_tasks_from_runtime(carnival_tasks_module: str) -> Dict[str, Type[Task]]:
    tasks: Dict[str, Type[Task]] = {}

    for task_class in task_subclasses(Task):
        task_full_name = get_task_full_name(carnival_tasks_module, task_class)
        tasks[task_full_name] = task_class

    return tasks


def get_tasks(carnival_tasks_module: str, for_completion: bool = False) -> Dict[str, Type[Task]]:
    sys.path.insert(0, os.getcwd())
    from carnival import internal_tasks  # noqa
    import_tasks_file(carnival_tasks_module, silent=for_completion)
    return get_tasks_from_runtime(carnival_tasks_module)



from dataclasses import dataclass
from typing import List, Union, Any, Type, Optional
import abc

import re

from carnival import global_context, Step

from carnival.host import Host


def _underscore(word: str) -> str:
    # https://github.com/jpvanhal/inflection/blob/master/inflection.py
    word = re.sub(r"([A-Z]+)([A-Z][a-z])", r'\1_\2', word)
    word = re.sub(r"([a-z\d])([A-Z])", r'\1_\2', word)
    word = word.replace("-", "_")
    return word.lower()


@dataclass
class TaskResult:
    """
    Возвращается вызовом метода Task.step
    """
    host: Host
    step: Step
    result: Any


class Task:
    """
    Задача это единица выполнения одного или несколькоих шагов на определенных хостах.

    Именование задач.

    Полное имя задачи состоит из двух частей. <module_name>.<name>.
    carnival автоматически генерирует имена задач из этих частей, но есть возможность управлять этим вручную,
    используя два атрибута класса Task.

    name: название задачи. если не определено имя будет сгенерировано автоматически.
    module_name: имя модуля. если назначить пустую строку, полное имя будет включать только название задачи.
    """

    # Имя задачи
    name: str = ""
    module_name: Optional[str] = None

    @classmethod
    def get_name(cls) -> str:
        return cls.name if cls.name else _underscore(cls.__name__)

    def __init__(self, dry_run: bool):
        self.dry_run = dry_run

    def call_task(self, task_class: Type['Task']):
        """
        Запустить другую задачу
        Возвращает результат работы задачи
        """
        return task_class(dry_run=self.dry_run).run()

    def step(self, steps: Union[Step, List[Step]], hosts: Union[Host, List[Host]]) -> List[TaskResult]:
        """
        Запустить шаг(и) на хост(ах)
        Возвращает объект TaskResult для получения результатов работы каждого шага на каждом хосте
        """

        if not isinstance(steps, list) and not isinstance(steps, tuple):
            steps = [steps, ]

        if not isinstance(hosts, list) and not isinstance(hosts, tuple):
            hosts = [hosts, ]

        results = []

        for host in hosts:
            with global_context.SetContext(host):
                for step in steps:
                    step_name = _underscore(step.__class__.__name__)
                    print(f"💃💃💃 Running {self.get_name()}:{step_name} at {host}")
                    if not self.dry_run:
                        r = TaskResult(
                            host=host,
                            step=step,
                            result=step.run_with_context(host=host),
                        )
                        results.append(r)
        return results

    @abc.abstractmethod
    def run(self):
        """
        Реализация выполнения задачи
        """
        raise NotImplementedError


class SimpleTask(abc.ABC, Task):
    """
    Запустить шаги `self.steps` на хостах `self.hosts`
    """

    hosts: List[Host]
    steps: List[Step]

    def run(self):
        self.step(
            steps=self.steps,
            hosts=self.hosts,
        )



import os

from jinja2 import Environment, FileSystemLoader, ChoiceLoader, PrefixLoader, PackageLoader

from carnival import global_context
from carnival.secrets_manager import secrets_storage
from carnival.plugins import discover_plugins

"""
Initialize loader on current working dir and plugin modules
"""
j2_env = Environment(loader=ChoiceLoader([
    FileSystemLoader(os.getcwd()),
    PrefixLoader({x: PackageLoader(x, package_path="") for x in discover_plugins().keys()}),
]))


def render(template_path: str, **context) -> str:
    template = j2_env.get_template(template_path)
    return template.render(
        conn=global_context.conn,
        connected_host=global_context.host,
        host_context=global_context,
        secrets=secrets_storage,
        **context,
    )



from carnival import global_context


def _run_command(command: str, **kwargs):
    return global_context.conn.run(command, **kwargs)


def run(command: str, **kwargs):
    return _run_command(command, **kwargs)


def pty(command: str, **kwargs):
    return run(command, pty=True, **kwargs)



from carnival import tasks_loader


def test_load_tasks_file():
    initial_tasks = tasks_loader.get_tasks_from_runtime("")

    tasks_loader.import_tasks_file("not_exists_file123_", silent=True)
    assert tasks_loader.get_tasks_from_runtime("") == initial_tasks

    tasks_loader.import_tasks_file("testdata.carnival_file_test", silent=False)
    tasks = tasks_loader.get_tasks_from_runtime("")
    assert len(tasks) == len(initial_tasks) + 1



from io import StringIO

from fabric.transfer import Transfer, Result  # type:ignore
from patchwork import transfers  # type:ignore

from carnival import global_context
from carnival.templates import render


def rsync(source, target, exclude=(), delete=False, strict_host_keys=True, rsync_opts="--progress -pthrvz", ssh_opts=''):
    # https://fabric-patchwork.readthedocs.io/en/latest/api/transfers.html#patchwork.transfers.rsync
    return transfers.rsync(
        c=global_context.conn,
        source=source,
        target=target,
        exclude=exclude,
        delete=delete,
        strict_host_keys=strict_host_keys,
        rsync_opts=rsync_opts,
        ssh_opts=ssh_opts,
    )


def get(remote: str, local: str, preserve_mode: bool = True) -> Result:
    # http://docs.fabfile.org/en/2.5/api/transfer.html#fabric.transfer.Transfer.get
    t = Transfer(global_context.conn)
    return t.get(remote=remote, local=local, preserve_mode=preserve_mode)


def put(local: str, remote: str, preserve_mode: bool = True) -> Result:
    # http://docs.fabfile.org/en/2.5/api/transfer.html#fabric.transfer.Transfer.put
    t = Transfer(global_context.conn)
    return t.put(local=local, remote=remote, preserve_mode=preserve_mode)


def put_template(template_path: str, remote: str, **context) -> Result:
    filestr = render(template_path=template_path, **context)
    t = Transfer(global_context.conn)
    return t.put(local=StringIO(filestr), remote=remote, preserve_mode=False)



from typing import Union

from fabric import Connection  # type:ignore
from invoke import Context  # type:ignore

from carnival.host import Host


# noinspection PyTypeChecker
conn: Union[Connection, Context] = None
# noinspection PyTypeChecker
host: Host = None  # type:ignore


class SetContext:
    def __init__(self, h: Host):
        self.host = h

    def __enter__(self):
        global conn
        global host

        assert host is None, f"Cannot set context, while other context active: {host}"
        assert conn is None, f"Cannot set context, while other context active: {conn}"

        conn = self.host.connect()
        host = self.host

    def __exit__(self, exc_type, exc_val, exc_tb):
        global conn
        global host
        conn = None
        host = None



import pytest

from carnival import cmd, global_context


@pytest.mark.slow
def test_apt_install(suspend_capture, ubuntu_ssh_host):
    with suspend_capture:
        with global_context.SetContext(ubuntu_ssh_host):
            assert cmd.fs.is_file_exists("/usr/bin/mc") is False
            assert cmd.apt.install("mc", hide=True) is True
            assert cmd.apt.install("mc", hide=True) is False
            assert cmd.fs.is_file_exists("/usr/bin/mc") is True
            cmd.apt.remove("mc", hide=True)


@pytest.mark.slow
def test_apt_install_multiple(suspend_capture, ubuntu_ssh_host):
    with suspend_capture:
        with global_context.SetContext(ubuntu_ssh_host):
            assert cmd.fs.is_file_exists("/usr/bin/mc") is False
            assert cmd.fs.is_file_exists("/usr/bin/htop") is False
            assert cmd.apt.install_multiple("htop", "mc", hide=True) is True
            assert cmd.apt.install_multiple("htop", "mc", hide=True) is False
            assert cmd.fs.is_file_exists("/usr/bin/mc") is True
            assert cmd.fs.is_file_exists("/usr/bin/htop") is True
            cmd.apt.remove("mc", "htop", hide=True)


@pytest.mark.slow
def test_apt_get_installed_version(suspend_capture, ubuntu_ssh_host):
    with suspend_capture:
        with global_context.SetContext(ubuntu_ssh_host):
            assert cmd.fs.is_file_exists("/usr/bin/mc") is False
            assert cmd.apt.get_installed_version("mc") is None

            cmd.apt.install("mc", hide=True)

            assert cmd.fs.is_file_exists("/usr/bin/mc") is True
            ver = cmd.apt.get_installed_version("mc")
            assert ver in cmd.apt.get_pkg_versions("mc")
            assert isinstance(ver, str)
            assert " " not in ver
            cmd.apt.remove("mc", hide=True)


@pytest.mark.slow
def test_apt_is_pkg_installed(suspend_capture, ubuntu_ssh_host):
    with suspend_capture:
        with global_context.SetContext(ubuntu_ssh_host):
            assert cmd.fs.is_file_exists("/usr/bin/mc") is False
            assert cmd.apt.is_pkg_installed("mc") is False

            cmd.apt.install("mc", hide=True)

            assert cmd.fs.is_file_exists("/usr/bin/mc") is True
            assert cmd.apt.is_pkg_installed("mc") is True
            ver = cmd.apt.get_installed_version("mc")
            assert cmd.apt.is_pkg_installed("mc", ver) is True
            assert cmd.apt.is_pkg_installed("mc", "3i2oxyom3x3 ") is False
            cmd.apt.remove("mc", hide=True)



import pytest

from carnival import secrets_manager
from carnival.secrets_manager import SecretGetter


def test_secrets_manager():
    with pytest.raises(NotImplementedError):
        SecretGetter().get_secret("TEST")


def test_secrets_manager_static():
    assert secrets_manager.secrets_storage == {}

    secrets_manager.secret("STATIC_SECRET", secrets_manager.Static("SECRET_VALUE"))
    assert secrets_manager.secrets_storage['STATIC_SECRET'] == 'SECRET_VALUE'

    secrets_manager.secrets_storage = {}


def test_secrets_manager_from_cli(mocker):
    assert secrets_manager.secrets_storage == {}

    mocker.patch('getpass.getpass', new=lambda p: "SECRET_VALUE")
    secrets_manager.secret("STATIC_SECRET", secrets_manager.FromCli())
    assert secrets_manager.secrets_storage['STATIC_SECRET'] == 'SECRET_VALUE'
    secrets_manager.secrets_storage = {}


def test_secrets_manager_from_env(mocker):
    assert secrets_manager.secrets_storage == {}
    mocker.patch('os.getenv', new=lambda v, d: None)
    secrets_manager.secret("STATIC_SECRET", secrets_manager.FromEnv(default="123", required=False))
    assert secrets_manager.secrets_storage['STATIC_SECRET'] == '123'
    secrets_manager.secrets_storage = {}

    with pytest.raises(secrets_manager.SecretGetError):
        assert secrets_manager.secrets_storage == {}
        mocker.patch('os.getenv', new=lambda v, d: None)
        secrets_manager.secret("STATIC_SECRET", secrets_manager.FromEnv(default="123", required=True))
        secrets_manager.secrets_storage = {}

    assert secrets_manager.secrets_storage == {}
    mocker.patch('os.getenv', new=lambda v, d: "SECRET_VALUE")
    secrets_manager.secret("STATIC_SECRET", secrets_manager.FromEnv(required=True))
    assert secrets_manager.secrets_storage['STATIC_SECRET'] == 'SECRET_VALUE'
    secrets_manager.secrets_storage = {}



from setuptools import setup, find_packages


requirements = [
    'Fabric==2.5.0',
    'invoke==1.4.0',
    'patchwork==1.0.1',
    'Jinja2==2.10.3',
    "Click==7.0",
]

VERSION = "0.10"

setup(
    name='carnival',
    version=VERSION,
    packages=find_packages(),
    url='https://github.com/carnival-org/carnival',
    license='MIT',
    author='a1fred',
    author_email='demalf@gmail.com',
    description='Fabric-based software provisioning tool',
    entry_points={
        'console_scripts': [
            'carnival = carnival.cli:main',
        ],
    },
    classifiers=[
        'Environment :: Console',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3',
    ],
    install_requires=requirements,
    test_suite="tests",
)



import os
from typing import Iterable

import click

from carnival.tasks_loader import get_tasks

carnival_tasks_module = os.getenv("CARNIVAL_TASKS_MODULE", "carnival_tasks")


def is_completion_script(complete_var: str) -> bool:
    return os.getenv(complete_var, None) is not None


task_types: dict = {}


def main():
    global task_types

    complete_var = os.getenv("COMPLETE_VAR", "_CARNIVAL_COMPLETE")
    task_types = get_tasks(
        carnival_tasks_module=carnival_tasks_module,
        for_completion=is_completion_script(complete_var)
    )

    @click.command()
    @click.option('-d', '--dry_run', is_flag=True, default=False, help="Simulate run")
    @click.argument('tasks', required=True, type=click.Choice(task_types.keys()), nargs=-1)
    def cli(dry_run: bool, tasks: Iterable[str]):
        for task in tasks:
            task_types[task](dry_run=dry_run).run()

    cli(complete_var=complete_var)



import inspect
from typing import List, Dict, Any
import abc

from carnival.host import Host

from carnival.secrets_manager import secrets_storage


def _get_arg_names(fn) -> List[str]:
    arg_names: List[str] = []
    spec = inspect.getfullargspec(fn)
    arg_names += spec.args
    arg_names += spec.kwonlyargs
    return arg_names


def _build_kwargs(fn, context: Dict[str, Any]):
    arg_names: List[str] = _get_arg_names(fn)
    kwargs = {}

    for context_name, context_val in context.items():
        if context_name in arg_names:
            kwargs[context_name] = context_val
    return kwargs


class Step:
    """
    Объект, предназначенный для выполнения группы комманд с какой-то целью.
    Вызывается из класса `carnival.Task` для выполнения комманд (`carnival.cmd`) на определенных хостах.

    Может требовать наличие определенных контекстных переменных для работы, указав их в аргументах метода `run`.
    Может вернуть значение для дальнейшего использования.

    В следующем примере переменная `disk_name` будет передана в run, а `install` пропущена.

    >>> host = Host(
    >>>     #  Адрес
    >>>     "1.2.3.4",
    >>>
    >>>     # Контекст хоста
    >>>     disk_name="/dev/sda1", install=['nginx', 'htop', ]
    >>> )
    >>> ...
    >>> class DiskUsage(Step):
    >>>     def run(self, disk_name: str):
    >>>         ...

    """
    def __init__(self, **context):
        """
        :param context: Переменные контекста, назначенные при вызове Шага
        """
        self.context = context

    def _build_context(self, host: Host) -> Dict[str, Any]:
        run_context = {'secrets': secrets_storage, 'host': host}
        if host.context:
            run_context.update(host.context)
        if self.context:
            run_context.update(self.context)
        return run_context

    def run_with_context(self, host: Host):
        context = self._build_context(host)
        kwargs = _build_kwargs(self.run, context)
        return self.run(**kwargs)

    @abc.abstractmethod
    def run(self, **kwargs):
        """
        Метод который нужно определить для выполнения комманд

        :param kwargs: Автоматические подставляемые переменные контекста
        """
        raise NotImplementedError



import pytest

from carnival import Step, Host


def test_step_abc(mocker):
    spy = mocker.spy(Step, 'run')
    with pytest.raises(NotImplementedError):
        Step(another="context1").run_with_context(Host("local", add="context"))

    spy.assert_called_once()


def test_step_context_overload(mocker):
    class NoopStep(Step):
        def run(self, another=None):
            pass

    # Waiting step context
    noop_step = NoopStep(another="context1")
    spy = mocker.spy(noop_step, 'run')
    noop_step.run_with_context(Host("local", add="context"))
    spy.assert_called_once_with(another='context1')

    # Waiting no context
    noop_step = NoopStep()
    spy = mocker.spy(noop_step, 'run')
    noop_step.run_with_context(Host("local", add="context"))
    spy.assert_called_once_with()

    # Waiting host context
    noop_step = NoopStep()
    spy = mocker.spy(noop_step, 'run')
    noop_step.run_with_context(Host("local", another="host_context"))
    spy.assert_called_once_with(another="host_context")

    # Waiting step and host context, step context overloading step context
    noop_step = NoopStep(another="context1")
    spy = mocker.spy(noop_step, 'run')
    noop_step.run_with_context(Host("local", another="host_context"))
    spy.assert_called_once_with(another="context1")



from carnival import cmd


def daemon_reload():
    cmd.cli.run(f"sudo systemctl --system daemon-reload")


def start(service_name: str, reload_daemon=False):
    if reload_daemon:
        daemon_reload()
    cmd.cli.run(f"sudo systemctl start {service_name}")


def stop(service_name: str, reload_daemon=False):
    if reload_daemon:
        daemon_reload()
    cmd.cli.run(f"sudo systemctl stop {service_name}")


def restart(service_name: str):
    cmd.cli.run(f"sudo systemctl restart {service_name}")


def enable(service_name: str, reload_daemon=False, start_now=True):
    if reload_daemon:
        daemon_reload()
    cmd.cli.run(f"sudo systemctl enable {service_name}")
    if start_now:
        start(service_name)


def disable(service_name: str, reload_daemon=False, stop_now=True):
    if reload_daemon:
        daemon_reload()
    cmd.cli.run(f"sudo systemctl disable {service_name}")
    if stop_now:
        stop(service_name)
